import 'dart:async';
import 'dart:convert';

import 'package:apex_socket/cipher/crypto.dart';
import 'package:flutter/cupertino.dart';
import 'package:logging/logging.dart';
import 'package:socket_io_client/socket_io_client.dart' as IO;

export 'package:socket_io_client/socket_io_client.dart';

typedef dynamic EventHandler<T>(T data);

class ApexSocket extends ChangeNotifier {
  final log = Logger('ApexSocket');

  IO.Socket socket;

  int _reconnectDelay = 2;

  /// Check if socket object is created and connected as well :)
  bool get isConnected => socket != null && socket.connected;

  /// Reconnection delay in seconds
  int get reconnectionDelay => _reconnectDelay;

  final int port;
  final String url;

  /// Just works on encrypt = true
  final String prefixData;
  final String namespace;
  final IO.OptionBuilder options;
  final bool encrypt;
  final String _secretKey;
  final String _publicKey;

  ApexSocket(this.url, this.port, this.options,
      {this.prefixData,
      this.namespace,
      this.encrypt = false,
      String secretKey,
      String publicKey})
      : _secretKey = secretKey,
        _publicKey = publicKey,
        assert(
            (encrypt && secretKey != null && publicKey != null) || (!encrypt)),
        assert(url != null && Uri.parse(url).isAbsolute,
            '$url must be a valid url.'),
        assert(port == null || (port >= -1 && port <= 65535),
            '$port must be a number between -1 and 65535. or null.'),
        assert(options != null, 'Options must not be null');

  void connect() {
    if (this.socket != null) this.socket.connect();
  }

  void createSocket({
    bool recreate = false,
    EventHandler onConnect,
    EventHandler onConnectError,
    EventHandler onError,
    EventHandler onDisconnect,
    EventHandler onReconnect,
  }) {
    assert(options != null, 'Options must not be null');
    if (socket == null || recreate) {
      if (recreate) if (this.socket != null) this.socket.destroy();
      IO.Socket socket = IO.io(
          '$url${port != null ? ':$port' : ''}${namespace != null ? '/$namespace' : ''}',
          options
              .setReconnectionDelay(_reconnectDelay)
              .setTransports(['websocket']).build());

      socket.onConnect((data) {
        _reconnectDelay = 2;
        notifyListeners();
        if (onConnect != null) return onConnect(data);
      });

      socket.onConnectError(onConnectError);

      socket.onError(onError);

      socket.onDisconnect((data) {
        notifyListeners();
        if (onDisconnect != null) return onDisconnect(data);
      });

      socket.onReconnect(onReconnect);

      socket.onReconnecting((data) {
        if (_reconnectDelay < 64) {
          _reconnectDelay *= 2;
        } else {
          _reconnectDelay = 64;
        }
        socket.io.options =
            options.setReconnectionDelay(_reconnectDelay).build();
        notifyListeners();
      });

      this.socket = socket;
    }

    if (!this.socket.connected)
      this.socket.connect();
    else
      log.info('already connected');
  }

  void emit(String event, {String data, bool enc}) {
    bool shouldEncrypt = enc == null ? encrypt : enc;

    if (shouldEncrypt) {
      final crypto = Crypto(_secretKey, _publicKey);
      data = (prefixData ?? '') + ':' + crypto.encrypt(data);
    } else {
      data = (prefixData ?? '') + ':' + base64Encode(utf8.encode(data));
    }
    this.socket.connect().emit(event, data);
  }

  void emitWithAck(String event, String data,
      {Function(String data) ack, bool enc}) {
    final shouldEncrypt = ((enc == null) ? encrypt : enc);

    Crypto crypto;
    if (shouldEncrypt) {
      crypto = Crypto(_secretKey, _publicKey);
      data = (prefixData ?? '') + ':' + crypto.encrypt(data);
    } else {
      data = (prefixData ?? '') + ':' + base64Encode(utf8.encode(data));
    }
    this.socket.connect().emitWithAck(event, data, ack: (m) {
      if (shouldEncrypt) {
        try {
          final decryptedMessage = crypto.decrypt(m.toString());
          ack(decryptedMessage);
        } catch (ignore) {
          ack(null);
        }
      } else
        ack(m.toString());
    });
  }

  Future emitWithAckC(String event, String data, [bool enc]) {
    final shouldEncrypt = ((enc == null) ? encrypt : enc);

    Completer completer = Completer();

    Crypto crypto;
    if (shouldEncrypt) {
      crypto = Crypto(_secretKey, _publicKey);
      data = (prefixData ?? '') + ':' + crypto.encrypt(data);
    } else {
      data = (prefixData ?? '') + ':' + base64Encode(utf8.encode(data));
    }
    this.socket.connect().emitWithAck(event, data, ack: (m) {
      if (shouldEncrypt) {
        try {
          final decryptedMessage = crypto.decrypt(m.toString());
          if (!completer.isCompleted) completer.complete(decryptedMessage);
        } catch (e) {
          if (!completer.isCompleted)
            completer.completeError(NullResponseError(e.toString()));
        }
      } else {
        if (!completer.isCompleted) completer.complete(m.toString());
      }
    });

    this.socket.on('connect_error', (data) {
      if (!completer.isCompleted)
        completer.completeError(ConnectError(data.toString()));
    });

    return completer.future;
  }
}

class NullResponseError implements Exception {
  final String message;

  NullResponseError(this.message);

  @override
  String toString() {
    return this.message;
  }
}

class ConnectError implements Exception {
  final String message;

  ConnectError(this.message);

  @override
  String toString() {
    return this.message;
  }
}
